import { type SceneState, type SceneType } from "./base"
import type { Game } from "../game"
import type { ChoiceSceneDefinition, TextSceneDefinition } from "../../shared/script/types"
import type { ChoiceSceneEvent } from "../../shared/scene-types/choice"
import type { SessionId } from "../session"
import { getSuggestedInteractionId, type InteractionSceneEvent } from "../../shared/scene-types/interaction"
import type { SuggestedInteraction } from "../../shared/mutations"

export const ChoiceSceneType: SceneType<ChoiceSceneDefinition, ChoiceSceneState> = {
    id: "choice",
    createState(game: Game, definition: ChoiceSceneDefinition): ChoiceSceneState {
        return new ChoiceSceneState(game, definition)
    }
}

class ChoiceSceneState implements SceneState<ChoiceSceneEvent> {
    private optionsById: Map<string, { votes: Set<SessionId> }> = new Map()
    private votedOptionIdBySessionId: Map<SessionId, string> = new Map()
    private isVotingConcluded: boolean = false

    constructor(private game: Game, private definition: ChoiceSceneDefinition) {
        definition.options.forEach(o => this.optionsById.set(o.id, { votes: new Set() }))
    }

    getConnectionEvents(): ChoiceSceneEvent[] {
        const events: ChoiceSceneEvent[] = []
        events.push(this.getVotesChangedEvent())

        if (this.isVotingConcluded) events.push({ type: "voting-concluded" })
        return events
    }

    removePlayer(sessionId: SessionId) {
        const votedOptionId = this.votedOptionIdBySessionId.get(sessionId)
        if (votedOptionId !== undefined) {
            const { votes } = this.optionsById.get(votedOptionId)!
            votes.delete(sessionId)
            this.emitVotesChanged()
        }

        this.votedOptionIdBySessionId.delete(sessionId)
    }

    private emit(event: ChoiceSceneEvent) {
        this.game.eventBus.emit("player-broadcast", { type: "scene-event", event })
    }

    private getVotesChangedEvent(): ChoiceSceneEvent {
        return { type: "votes-changed", votesByOptionId: Object.fromEntries(this.optionsById.entries().map(([id, { votes }]) => ([id, votes.size]))) }
    }

    private emitVotesChanged() {
        this.emit(this.getVotesChangedEvent())
    }

    setVote(optionId: string | null, sessionId: SessionId) {
        const currentVotedOptionId = this.votedOptionIdBySessionId.get(sessionId)
        if (currentVotedOptionId !== undefined) {
            if (optionId === currentVotedOptionId) return
            const { votes } = this.optionsById.get(currentVotedOptionId)!
            votes.delete(sessionId)
            this.votedOptionIdBySessionId.delete(sessionId)
        }

        if (optionId !== null) {
            const { votes } = this.optionsById.get(optionId)!
            votes.add(sessionId)
            this.votedOptionIdBySessionId.set(sessionId, optionId)
        }

        this.emitVotesChanged()
    }

    concludeVoting() {
        if (this.isVotingConcluded) return
        this.isVotingConcluded = true
        this.emit({ type: "voting-concluded"})
    }

    restartVoting() {
        this.isVotingConcluded = false
        this.optionsById.forEach(({ votes }) => votes.clear())
        this.votedOptionIdBySessionId.clear()
        this.emitVotesChanged()
        this.emit({ type: "voting-restarted" })
    }
}